@@ -162,9 +162,8 @@ findPSname P1C(char *, name)
  *   This routine adds an entry.
  */
 void
-add_entry P7C(char *, TeXname, char *, PSname, char *, Fontfile,
-             char *, Vectfile, char *, specinfo, char *, downloadinfo,
-             boolean, partial)
+add_entry P6C(char *, TeXname, char *, PSname, char *, Fontfile,
+	      char *, Vectfile, char *, specinfo, char *, downloadinfo)
 {
    struct resfont *p ;
    int h ;
@@ -181,7 +180,6 @@ add_entry P7C(char *, TeXname, char *, PSname, char *, Fontfile,
       p->downloadheader = downloadinfo ;
    else
       p->downloadheader = 0 ;
-   p->partialdownload = partial ;
    h = hash(TeXname) ;
    p->next = reshash[h] ;
    p->sent = 0 ;
@@ -257,10 +255,10 @@ residentfont P1C(register fontdesctype *, curfnt)
             add_header(q) ;
             *cp++ = ' ' ;
          } else {
-            if (strstr(q,".pfa")||strstr(q,".pfb")||
+/*          if (strstr(q,".pfa")||strstr(q,".pfb")||
                 strstr(q,".PFA")||strstr(q,".PFB"))
-               break ;       /* We don't add PS font file here */
-            else {
+               break ;
+            else */ {
               add_header(q) ;
               break;
             }
@@ -546,10 +544,15 @@ case 'M' :
 case 'o' :
 	 if (!oname_option) {
            oname = configstring(was_inline+1, 1) ;
+#if defined(MSDOS) || defined(OS2)
            if (*oname && oname[strlen(oname)-1] == ':')
               sendcontrolD = 1 ; /* if we send to a device, *we* are spooler */
+#endif
 	 }
          break ;
+case 'F' :
+         sendcontrolD = (was_inline[1] != '0') ;
+         break ;
 case 'O' :
          p = was_inline + 1 ;
          handlepapersize(p, &hoff, &voff) ;
@@ -741,7 +744,7 @@ case 'e' :
 case 'q' : case 'Q' :
          quiet = (was_inline[1] != '0') ;
          break ;
-case 'f' : case 'F' :
+case 'f' :
          filter = (was_inline[1] != '0') ;
 	 if (filter)
 	   oname = "";
@@ -862,13 +865,13 @@ void getpsinfo P1C(char *, name)
                        p++;
                        encoding_p = true;
                      }
-                     
+                     p++ ;
                      /* skip whitespace after < */
-                     while (p[1] == ' ' || p[1] == '\t')
+                     while (*p && *p <= ' ')
                        p++;
                      
                      /* save start of header name */
-                     hdr_name = p + 1 ;
+                     hdr_name = p ;
 
                   } else if (TeXname) /* second regular word on line? */
                      PSname = p ;
@@ -895,19 +898,15 @@ void getpsinfo P1C(char *, name)
                         /* (SPQR) if it is a reencoding, pass on to
                            FontPart, and download as usual */
                         Vectfile = downloadinfo = hdr_name;
-		     } else if 	(!STREQ (suffix, "pfa")
-				 && !STREQ (suffix, "pfb")
-				 && !STREQ (suffix, "PFA")
-				 && !STREQ (suffix, "PFB")) {
-                       /* No special suffix. check if tagged <<*/
-                       if (nopartial_p) /* its a font after all */
-			 {  Fontfile = hdr_name;}
-                       else 
-                         {  downloadinfo = hdr_name; }
-		     }
-                     else {
-                        /* its a font suffix. */
-                        Fontfile = hdr_name;
+                     } else if (nopartial_p) {
+                        downloadinfo = hdr_name ;
+		     } else if 	(STREQ (suffix, "pfa")
+				 || STREQ (suffix, "pfb")
+				 || STREQ (suffix, "PFA")
+				 || STREQ (suffix, "PFB")) {
+			Fontfile = hdr_name;
+		     } else {
+                        downloadinfo = hdr_name;
                      }
                   }
                }
@@ -928,7 +927,7 @@ void getpsinfo P1C(char *, name)
                specinfo = newstring(specbuf) ;
                downloadinfo = newstring(downbuf) ;
                add_entry(TeXname, PSname, Fontfile, Vectfile,
-                         specinfo, downloadinfo, !nopartial_p) ;
+                         specinfo, downloadinfo) ;
             }
    	 }
       }
