@@ -65,33 +65,34 @@ typedef	struct direct	struct_dirent;
 #endif
 #endif /* no <dirent.h> */
 
+#ifdef HAVE_POLL
+#include <poll.h>
+#endif
+
 #ifndef atof
 double	atof();
 #endif
 char	*getenv();
 
+/* <sys/types.h> is already included.  */
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
+
 #ifndef WIFSTOPPED
-#define	WIFSTOPPED(status)	(((status) & 0377) == 0177)
+#define WIFSTOPPED(stat_val)    (((stat_val) & 0377) == 0177)
 #endif
 
 #ifndef WIFSIGNALED
-#define	WIFSIGNALED(status)	(((status) & 0377) != 0)
+#define WIFSIGNALED(stat_val)   (((stat_val) & 0377) != 0)
 #endif
 
 #ifndef WTERMSIG
-#ifndef USE_UNION_WAIT
-#define	WTERMSIG(status)	((status) & 0177)
-#else
-#define	WTERMSIG(status)	((status).w_termsig)
-#endif
+#define WTERMSIG(stat_val)      ((stat_val) & 0177)
 #endif
 
 #ifndef WEXITSTATUS
-#ifndef USE_UNION_WAIT
-#define	WEXITSTATUS(status)	(((status) >> 8) & 0377)
-#else
-#define	WEXITSTATUS(status)	((status).w_retcode)
-#endif
+#define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
 #endif
 
 #ifndef SIGCHLD
@@ -175,6 +176,7 @@ char	*getenv();
 
 #else /* KPATHSEA */
 
+#include <kpathsea/c-auto.h>
 #include <kpathsea/config.h>
 #include <kpathsea/c-errno.h>
 #include <kpathsea/c-ctype.h>
@@ -184,77 +186,72 @@ char	*getenv();
 #include <kpathsea/tex-file.h>
 #include <kpathsea/tex-make.h>
 #include <signal.h>
-#if !defined(_AMIGA) && !defined(WIN32)
+#ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
 #include <fcntl.h>
 #include <setjmp.h>
 
 #ifndef HAVE_STRCHR
-#define	strchr	index
+#define strchr  index
 #endif
 
 #ifndef HAVE_STRRCHR
-#define	strrchr	rindex
+#define strrchr rindex
+#endif
+
+#if defined(HAVE_POLL) && !defined(HAVE_POLL_H)
+#undef HAVE_POLL
 #endif
 
 /* Add two new flags to kpathsea_debug for debugging gsftopk */
 #define	GSPK_DEBUG_BITMAP	(KPSE_LAST_DEBUG + 1)
 #define	GSPK_DEBUG_PK		(KPSE_LAST_DEBUG + 2)
 
-#ifdef WIN32
-#include <process.h>
-#define	spawnlp	_spawnlp
-#define	sleep	Sleep
-#define	WIFSIGNALED(status)	(((status) & ((DWORD)0x0F0000000L)) != 0)
-#define	WEXITSTATUS(status)	(((status) >> 8) & 0x0FF)
-#else
-#ifndef WIFSTOPPED
-#define	WIFSTOPPED(status)	(((status) & 0377) == 0177)
-#endif
-
-#ifndef WIFSIGNALED
-#define	WIFSIGNALED(status)	(((status) & 0377) != 0)
-#endif
-
-#ifndef WTERMSIG
-#ifndef USE_UNION_WAIT
-#define	WTERMSIG(status)	((status) & 0177)
-#else
-#define	WTERMSIG(status)	((status).w_termsig)
-#endif
-#endif
-
-#ifndef WEXITSTATUS
-#ifndef USE_UNION_WAIT
-#define	WEXITSTATUS(status)	(((status) >> 8) & 0377)
-#else
-#define	WEXITSTATUS(status)	((status).w_retcode)
-#endif
-#endif
-
-#endif /* WIN32 */
-
 #ifdef HAVE_SYS_WAIT_H
 #include <sys/wait.h>
 #endif
 
+#ifdef WIN32
+
+#include <process.h>
+#define	spawnlp	_spawnlp
+#define	sleep	Sleep
+#define	WIFSIGNALED(stat_val)	(((stat_val) & ((DWORD)0x0F0000000L)) != 0)
+#define	WEXITSTATUS(stat_val)	(((stat_val) >> 8) & 0x0FF)
+
+#else /* not WIN32 */
+
+#ifndef WEXITSTATUS
+#define WEXITSTATUS(stat_val)   ((unsigned)(stat_val) >> 8)
+#endif
+
+#ifndef WIFSTOPPED
+#define WIFSTOPPED(stat_val)    (((stat_val) & 0377) == 0177)
+#endif
+
+#ifndef WIFSIGNALED
+#define WIFSIGNALED(stat_val)   (((stat_val) & 0377) != 0)
+#endif
+
+#ifndef WTERMSIG
+#define WTERMSIG(stat_val)      ((stat_val) & 0177)
+#endif
+
+#endif /* not WIN32 */
+
 #ifdef HAVE_VFORK_H
 #include <vfork.h>
 #else
 #define vfork fork
 #endif
 
-#endif /* KPATHSEA */
-
-#if defined(HAVE_POLL) && !defined(HAVE_POLL_H)
-#undef HAVE_POLL
-#endif
-
 #ifdef HAVE_POLL
 #include <poll.h>
 #endif
 
+#endif /* KPATHSEA */
+
 #ifdef _AMIGA
 #include <proto/dos.h>
 #include <dos/dostags.h>
@@ -600,11 +597,7 @@ handle_sigterm P1C(int, signo)
 
 #define	gs_is_done	(gs_pid == 0)
 
-#ifdef USE_UNION_WAIT
-typedef	union wait	gsf_wait_t;
-#else
 typedef	int		gsf_wait_t;
-#endif
 
 static void
 wait_for_gs P1H(void)
@@ -2207,7 +2200,11 @@ Primary author of gsftopk: Paul Vojta; -k maintainer: K. Berry.");
 		/* render.ps */ searchpath,
 		PSname != NULL ? PSname : fontname,
 		dlstring != NULL ? dlstring : "", specinfo, dpistr, NULL);
-	    perror("gs");
+	    if (col != 0) {
+		putc('\n', stderr);
+		col = 0;
+	    }
+	    perror(gspath);
 	    exit(1);
 	}
 	if (gs_pid == -1) {
