@@ -18,14 +18,12 @@
 #define texpsheaderbopenin(f) \
     open_input (&(f), kpse_tex_ps_header_format, FOPEN_RBIN_MODE)
 
-extern integer fmlookup();
-extern boolean sharedsrc();
 extern void writezip();
-extern void writeimage();
+extern void writeimg();
 extern void dopdffont();
 extern void readconfig();
 extern void adjustcfgdimens();
-extern integer readimageinfo();
+extern integer readimg();
 extern void addimageref();
 extern void deleteimageref();
 extern void libpdffinish();
@@ -33,10 +31,21 @@ extern integer imagewidth();
 extern integer imageheight();
 extern integer imagexres();
 extern integer imageyres();
+extern boolean iscolorimage();
+extern boolean ispdfimage();
 extern integer cfgoutput();
 extern integer cfgcompresslevel();
 extern integer cfgdecimaldigits();
+extern integer cfgpkresolution();
 extern integer cfgpagewidth();
 extern integer cfgpageheight();
 extern integer cfghorigin();
 extern integer cfgvorigin();
+extern integer newvfpacket();
+extern unsigned char packetbyte();
+extern void pushpacketstate();
+extern void poppacketstate();
+extern void startpacket();
+extern integer fmlookup();
+extern boolean sharedsrc(int,int);
+extern void storepacket(int,int,int);
