@@ -36,11 +36,15 @@ notdef, "dotlessi", notdef, notdef, "lslash", "oslash", "oe", "germandbls",
 notdef, notdef, notdef, notdef
 };
 
-#define T1_BUF_SIZE   1024
+#define T1_BUF_SIZE   4096
 
 #define ENC_BUITIN    0
 #define ENC_STANDARD  1
 
+#define T1_TYPE_1     1
+#define T1_TYPE_2     2
+#define T1_TYPE_3     3
+
 typedef unsigned char byte;
 
 typedef struct {
@@ -56,15 +60,18 @@ static unsigned short t1_c1 = 52845, t1_c2 = 22719;
 static unsigned short t1_cslen, t1_lenIV;
 static char t1_line[T1_BUF_SIZE], *t1_line_ptr, *cs_start;
 static boolean t1_in_eexec, t1_ispfa, t1_charstring;
+static long t1_block_length;
 static FILE *t1_file;
 
-
 #define T1_OPEN()       typeonebopenin(t1_file)
 #define T1_CLOSE()      xfclose(t1_file, filename)
 #define T1_GETCHAR()    xgetc(t1_file)
 #define T1_EOF()        feof(t1_file)
 #define T1_PREFIX(s)    (!strncmp(t1_line, s, strlen(s)))
 #define T1_PUTCHAR(c)   pdfout(c)
+#define T1_CHARSTRINGS() (T1_PREFIX("2 index /CharStrings") || \
+                         T1_PREFIX("dup /CharStrings") ||  \
+                         T1_PREFIX("/CharStrings"))
 
 #define T1_CHECK_EOF()                                     \
     if (T1_EOF())                                          \
@@ -76,6 +83,38 @@ static FILE *t1_file;
     t1_putline();                                          \
 } while (0)
 
+static void t1_check_pfa()
+{
+    int c = T1_GETCHAR();
+    if (c != 128)
+        t1_ispfa = true;
+    else 
+        t1_ispfa = false;
+    ungetc(c, t1_file);
+}
+
+static byte t1_getbyte()
+{
+    int c = T1_GETCHAR();
+    if (t1_ispfa)
+        return c;
+    if (t1_block_length == 0) {
+        if (c != 128)
+            FAIL("invalid marker");
+        c = T1_GETCHAR();
+        if (c == 3)
+            FAIL("unexpected end of file");
+        t1_block_length = T1_GETCHAR() & 0xff;
+        t1_block_length |= (T1_GETCHAR() & 0xff) << 8;
+        t1_block_length |= (T1_GETCHAR() & 0xff) << 16;
+        t1_block_length |= (T1_GETCHAR() & 0xff) << 24;
+        c = T1_GETCHAR();
+        T1_CHECK_EOF();
+    }
+    t1_block_length--;
+    return c;
+}
+
 static int hexval(int c)
 {
     if (c >= 'A' && c <= 'F')
@@ -93,8 +132,8 @@ static byte edecrypt(byte cipher)
     byte plain;
     if (t1_ispfa) {
         while (cipher == 10 || cipher == 13)
-            cipher = T1_GETCHAR();
-        cipher = (hexval(cipher) >> 4) + hexval(T1_GETCHAR());
+            cipher = t1_getbyte();
+        cipher = (hexval(cipher) << 4) + hexval(t1_getbyte());
     }
     plain = (cipher^(t1_dr >> 8));
     t1_dr = (cipher + t1_dr)*t1_c1 + t1_c2;
@@ -115,23 +154,6 @@ static byte eencrypt(byte plain)
     return cipher;
 }
 
-static void t1_marker(int c, int m) 
-{
-    int i;
-    while (c == 10 || c == 13) {
-        c = T1_GETCHAR();
-        T1_CHECK_EOF();
-    }
-    if (c != 128)
-        FAIL("invalid marker");
-    if (T1_GETCHAR() != m)
-        FAIL("invalid block type");
-    for (i = 0; i < 4; i++) {
-        T1_GETCHAR();
-        T1_CHECK_EOF();
-    }
-} 
-
 static boolean t1_suffix(char *s) 
 {
     char *s1 = t1_line_ptr - 1, 
@@ -152,7 +174,7 @@ static void t1_getline()
 restart:
     t1_line_ptr = t1_line;
     t1_cslen = 0;
-    c = T1_GETCHAR();
+    c = t1_getbyte();
     while (!T1_EOF()) {
         if (t1_in_eexec) 
             c = edecrypt(c);
@@ -170,11 +192,11 @@ restart:
             cs_start = t1_line_ptr;
             CHECK_BUF(t1_line_ptr - t1_line + l, T1_BUF_SIZE);
             while (l-- > 0) {
-                *t1_line_ptr++ = edecrypt(T1_GETCHAR());
+                *t1_line_ptr++ = edecrypt(t1_getbyte());
                 T1_CHECK_EOF();
             }
         }
-        c = T1_GETCHAR();
+        c = t1_getbyte();
     }
     APPEND_EOL(t1_line_ptr, t1_line, T1_BUF_SIZE);
     if (t1_line_ptr - t1_line <= 1)
@@ -241,6 +263,85 @@ void t1_scan_param()
     key->value.i = atoi(p);
 }
 
+void t1_builtin_enc()
+{
+    int i, end_encoding = 0, counter = 0;
+    char buf[T1_BUF_SIZE], newline[T1_BUF_SIZE], *r, *q;
+       /*
+        * At this moment "/Encoding" is the prefix of t1_line
+        * 
+        * We have two possible forms of Encoding vector. The first case is
+        * 
+        *     /Encoding [/a /b /c...] readonly def
+        * 
+        * and the second case can look like
+        * 
+        *     /Encoding 256 array 0 1 255 {1 index exch /.notdef put} for
+        *     dup 0 /x put
+        *     dup 1 /y put
+        *     ...
+        *     readonly def
+        */
+    if (T1_PREFIX("/Encoding [") || T1_PREFIX("/Encoding[")) { /* the first case */
+        r = strchr(t1_line, '[') + 1;
+        strncpy(newline, t1_line, r - t1_line);
+        newline[r - t1_line + 1] = 0;
+        if (*r == 32)
+            r++;
+        for(;;) {
+            while (*r == '/') {
+                for (q = buf, r++; *r != 32 && *r != 10 && *r != ']' && *r != '/'; *q++ = *r++);
+                *q = 0;
+                if (*r == 32)
+                    r++;
+                if (strcmp(buf, notdef) != 0 && pdfischarused(tex_font, counter))
+                    builtin_glyph_names[counter] = xstrdup(buf);
+                sprintf(strchr(newline, 0), "/%s", buf);
+                counter++;
+            }
+            if (*r != 10 && *r != '%')
+                if (strncmp(r, "] def", strlen("] def")) == 0 ||
+                    strncmp(r, "] readonly def", strlen("] readonly def")) == 0) {
+                    end_encoding = 1;
+                    strcat(newline, r);
+                }
+                else
+                    FAIL("a name or `] def' or `] readonly def' expected:\n%s" AND t1_line);
+            strcpy(t1_line, newline);
+            t1_line_ptr = strchr(t1_line, 0);
+            APPEND_EOL(t1_line_ptr, t1_line, T1_BUF_SIZE);
+            t1_putline();
+            if (end_encoding)
+                break;
+            T1_CHECK_EOF();
+            t1_getline();
+            r = t1_line;
+            *newline = 0;
+        }
+    }
+    else { /* the second case */
+        do {
+            t1_putline();
+            T1_CHECK_EOF();
+            t1_getline();
+            strncpy(newline, t1_line, strlen(t1_line) - 1);
+            /* truncate the end-of-line character */
+        } while
+        (strstr("/Encoding 256 array 0 1 255 {1 index exch /.notdef put} for", newline) != NULL ||
+         strstr("/Encoding 256 array 0 1 255{1 index exch/.notdef put}for", newline) != NULL);
+        do {
+            if (sscanf(t1_line, "dup %u%s put", &i, buf) == 2 &&
+                *buf == '/' && pdfischarused(tex_font, i)) {
+                builtin_glyph_names[i] = xstrdup(buf + 1); /* skip the slash */
+                t1_putline();
+            }
+            T1_CHECK_EOF();
+            t1_getline();
+        } while (!t1_suffix("def"));
+        t1_putline();
+    }
+}
+
 void cs_store()
 {
     char *p, *q, buf[T1_BUF_SIZE];
@@ -249,9 +350,13 @@ void cs_store()
     *q = 0;
     cs_ptr->name = xstrdup(buf + 1); /* don't store the slash */
     cs_ptr->len = t1_cslen;
-    cs_ptr->data = XTALLOC(cs_ptr->len + 9, byte);
+    cs_ptr->data = XTALLOC(cs_ptr->len + 8, byte);
     cs_ptr->used = false;
-    bcopy(cs_start - 4, cs_ptr->data, cs_ptr->len + 9);
+    memcpy(cs_ptr->data, cs_start - 4, cs_ptr->len + 6);
+    p = cs_start - 4 + cs_ptr->len + 6;
+    q = cs_ptr->data + cs_ptr->len + 6;
+    for (; *p != 10; *q++ = *p++);
+    *q = 10;
     cs_ptr++;
 }
 
@@ -261,7 +366,7 @@ static void cs_mark(char *name)
 {
     byte *data;
     int i, b;
-    long val1, val2, cs_len;
+    long val1=0, val2=0, cs_len;
     unsigned short cr;
     cs_entry *cs;
     for (cs = cs_tab; cs < cs_ptr; cs++) {
@@ -319,7 +424,7 @@ static void cs_mark(char *name)
 void writet1()
 {
     int i, size_pos, cs_dict_size, cs_count, encoding;
-    char buf[T1_BUF_SIZE], *save_line1, *save_line2, *p;
+    char buf[T1_BUF_SIZE], *save_line1, *save_line2, *p, *q;
     cs_entry *cs;
     filename = fm_cur->ff_name;
     packfilename(maketexstring(filename), getnullstr(), getnullstr());
@@ -334,53 +439,42 @@ void writet1()
     t1_er = 55665;
     t1_in_eexec = false;
     t1_charstring = false;
-    if ((i = T1_GETCHAR()) != 128)
-        t1_ispfa = true;
-    else
-        t1_marker(i, 1);
+    t1_charstring = false;
+    t1_block_length = 0;
+    t1_check_pfa();
     if (!is_included()) { /* scan parameters from font file */
         do {
             T1_CHECK_EOF();
             t1_getline();
             t1_scan_param();
         } while (!T1_PREFIX("currentfile eexec"));
-        if (!t1_ispfa) {
-            i = T1_GETCHAR();
-            t1_marker(i, 2);
-        }
         t1_in_eexec = true;
         for (i = 0; i < 4; i++) {
             T1_CHECK_EOF();
-            edecrypt(T1_GETCHAR());
+            edecrypt(t1_getbyte());
         }
         do {
             T1_CHECK_EOF();
             t1_getline();
             t1_scan_param();
-        } while (!(T1_PREFIX("2 index /CharStrings") ||
-             T1_PREFIX("dup /CharStrings") ||
-             T1_PREFIX("/Subrs")));
+        } while (!(T1_CHARSTRINGS() || T1_PREFIX("/Subrs")));
         TEX_PRINTF(">");
         T1_CLOSE();
         return;
-    } 
+    }
     if (!is_subsetted()) { /* include entire font */
-        pdfsaveoffset = pdfoffset;
+        pdfsaveoffset = pdfoffset();
         do {
             T1_CHECK_EOF();
             t1_getline();
             t1_scan_param();
             t1_putline();
         } while (!T1_PREFIX("currentfile eexec"));
-        t1_length1 = pdfoffset - pdfsaveoffset;
-        if (!t1_ispfa) {
-            i = T1_GETCHAR();
-            t1_marker(i, 2);
-        }
-        pdfsaveoffset = pdfoffset;
+        t1_length1 = pdfoffset() - pdfsaveoffset;
+        pdfsaveoffset = pdfoffset();
         t1_in_eexec = true;
         for (t1_line_ptr = t1_line, i = 0; i < 4; i++) {
-            edecrypt(T1_GETCHAR());
+            edecrypt(t1_getbyte());
             T1_CHECK_EOF();
             *t1_line_ptr++ = 0;
         }
@@ -390,33 +484,29 @@ void writet1()
             t1_getline();
             t1_scan_param();
             t1_putline();
-        } while (!(T1_PREFIX("2 index /CharStrings") ||
-                 T1_PREFIX("dup /CharStrings") ||
-                 T1_PREFIX("/Subrs")));
+        } while (!(T1_CHARSTRINGS() || T1_PREFIX("/Subrs")));
         t1_charstring = true;
         do {
             T1_CHECK_EOF();
             t1_getline();
             t1_putline();
         } while (!t1_suffix("mark currentfile closefile"));
-        t1_length2 = pdfoffset - pdfsaveoffset;
-        if (!t1_ispfa)
-            t1_marker(T1_GETCHAR(), 1);
-        pdfsaveoffset = pdfoffset;
+        t1_length2 = pdfoffset() - pdfsaveoffset;
+        pdfsaveoffset = pdfoffset();
         t1_in_eexec = false;
         t1_charstring = false;
         do {
             T1_CHECK_EOF();
             t1_getline();
             t1_putline();
-        } while (!t1_suffix("cleartomark"));
-        t1_length3 = pdfoffset - pdfsaveoffset;
+        } while (!T1_PREFIX("cleartomark"));
+        t1_length3 = pdfoffset() - pdfsaveoffset;
         TEX_PRINTF(">");
         T1_CLOSE();
         return;
     } 
     /* the rest is for partial downloading */
-    pdfsaveoffset = pdfoffset;
+    pdfsaveoffset = pdfoffset();
     T1_CHECK_EOF();
     t1_getline();
     while (!T1_PREFIX("/Encoding")) {
@@ -432,64 +522,43 @@ void writet1()
         else 
             FAIL("cannot subset font (unknown predefined encoding `%s')"
                  AND buf);
-        t1_putline(); /* write the predefined encoding */
     }
     else {
         encoding = ENC_BUITIN; /* the font has its own built-in encoding */
-        /*
-        T1_PRINTF("/Encoding 256 array\n0 1 255 {1 index exch /.notdef put} for\n");
-        */
-        t1_putline();
     }
-    if (is_reencoded()) { /*  we needn't read the built-in encoding */
-        /*
-        T1_PRINTF("readonly def\n");
-        */
-        if (encoding == ENC_BUITIN) do {
+    if (is_reencoded()) { /* we needn't read the built-in encoding */
+        while (!t1_suffix("def")) {
+            t1_putline();
             T1_CHECK_EOF();
             t1_getline();
-            t1_putline();
-        } while (!t1_suffix("readonly def"));
+        }
+        t1_putline();
     }
     else if (encoding == ENC_BUITIN) { /* we must read the built-in encoding */
         for (i = 0; i < MAX_CHAR_NUM; i++)
             builtin_glyph_names[i] = notdef;
-        do {
-            T1_CHECK_EOF();
-            t1_getline();
-            if (sscanf(t1_line, "dup %u%s put", &i, buf) == 2 &&
-                *buf == '/' && pdfischarused(tex_font, i)) {
-                builtin_glyph_names[i] = xstrdup(buf + 1); /* skip the slash */
-                t1_putline();
-            }
-        } while (!t1_suffix("readonly def"));
-        T1_PRINTF("readonly def\n");
+        t1_builtin_enc();
     }
+    else
+        t1_putline(); /* write the predefined encoding */
     do {
         T1_CHECK_EOF();
         t1_getline();
         t1_scan_param();
         t1_putline();
     } while (!T1_PREFIX("currentfile eexec"));
-    t1_length1 = pdfoffset - pdfsaveoffset;
-    if (!t1_ispfa) {
-        i = T1_GETCHAR();
-        t1_marker(i, 2);
-    }
-    pdfsaveoffset = pdfoffset;
+    t1_length1 = pdfoffset() - pdfsaveoffset;
+    pdfsaveoffset = pdfoffset();
     t1_in_eexec = true;
     for (t1_line_ptr = t1_line, i = 0; i < 4; i++) {
-        edecrypt(T1_GETCHAR());
+        edecrypt(t1_getbyte());
         T1_CHECK_EOF();
         *t1_line_ptr++ = 0;
     }
     t1_putline(); /* to put the first four bytes */
     T1_CHECK_EOF();
     t1_getline();
-    while (!(T1_PREFIX("2 index /CharStrings") ||
-             T1_PREFIX("dup /CharStrings") ||
-             T1_PREFIX("/Subrs")))
-    {
+    while (!(T1_CHARSTRINGS() || T1_PREFIX("/Subrs"))) {
         t1_scan_param();
         t1_putline();
         T1_CHECK_EOF();
@@ -500,8 +569,7 @@ void writet1()
         t1_putline();
         T1_CHECK_EOF();
         t1_getline();
-    } while (!(T1_PREFIX("2 index /CharStrings") ||
-             T1_PREFIX("dup /CharStrings")));
+    } while (!T1_CHARSTRINGS());
     size_pos = strstr(t1_line, "/CharStrings") +
         strlen("/CharStrings") + 1 - t1_line;
     cs_dict_size = atoi(t1_line + size_pos);
@@ -524,7 +592,7 @@ void writet1()
     for (i = 0; i < MAX_CHAR_NUM; i++)
         if (pdfischarused(tex_font, i))
             if (cur_glyph_names[i] == notdef)
-                WARN("character %d is mapped to %s" AND
+                WARN("character %i is mapped to %s" AND
                      i AND notdef);
             else
                 cs_mark(cur_glyph_names[i]);
@@ -545,8 +613,11 @@ void writet1()
         if (cs->used) {
             sprintf(t1_line, "/%s %u", cs->name, cs->len);
             p = strchr(t1_line, 0);
-            bcopy(cs->data, p, cs->len + 9);
-            p += cs->len + 9;
+            memcpy(p, cs->data, cs->len + 6);
+            p += cs->len + 6;
+            q = cs->data + cs->len + 6;
+            for (; *q != 10; *p++ = *q++);
+            *p++ = 10;
             t1_line_ptr = p;
             t1_putline();
         }
@@ -561,18 +632,16 @@ void writet1()
         t1_getline();
         t1_putline();
     }
-    t1_length2 = pdfoffset - pdfsaveoffset;
-    if (!t1_ispfa)
-        t1_marker(T1_GETCHAR(), 1);
-    pdfsaveoffset = pdfoffset;
+    t1_length2 = pdfoffset() - pdfsaveoffset;
+    pdfsaveoffset = pdfoffset();
     t1_in_eexec = false;
     t1_charstring = false;
-    while (!t1_suffix("cleartomark")) {
+    while (!T1_PREFIX("cleartomark")) {
         T1_CHECK_EOF();
         t1_getline();
         t1_putline();
     }
-    t1_length3 = pdfoffset - pdfsaveoffset;
+    t1_length3 = pdfoffset() - pdfsaveoffset;
     TEX_PRINTF(">");
     XFREE(cs_tab);
     T1_CLOSE();
