@@ -75,7 +75,7 @@ typedef unsigned short  TTF_F2DOT14;
 #define COPY_UFWORD()   PUT_UFWORD(GET_UFWORD())
 #define COPY_F2DOT14()  PUT_F2DOT14(GET_F2DOT14())
 
-#define MAC_GLYPH_MAX   258
+#define NMACGLYPHS      258
 #define TABDIR_OFF      12
 
 typedef struct {
@@ -100,16 +100,32 @@ typedef struct {
     short newindex;         /* new index of glyph in output file */
 } ttfenc_entry;
 
+typedef struct {
+    TTF_USHORT platform_id;
+    TTF_USHORT encoding_id;
+    TTF_USHORT language_id;
+    TTF_USHORT name_id;
+    TTF_USHORT length;
+    TTF_USHORT offset;
+} name_record;
+
+typedef struct {
+    TTF_USHORT platform_id;
+    TTF_USHORT encoding_id;
+    TTF_ULONG  offset;
+    TTF_USHORT format;
+} cmap_entry;
+
 static TTF_USHORT ntabs;
 static TTF_ULONG checksum;
 static TTF_USHORT upem;
 static TTF_FIXED post_format;
 static TTF_SHORT loca_format;
 static TTF_ULONG ttf_offset;
+static TTF_ULONG last_glyf_offset;
 static TTF_USHORT glyphs_count;
 static TTF_USHORT new_glyphs_count;
 static TTF_USHORT nhmtxs;
-static TTF_USHORT max_glyph_index;
 static TTF_USHORT new_ntabs;
 
 /* 
@@ -125,59 +141,29 @@ static TTF_USHORT new_ntabs;
 static glyph_entry *glyph_tab;
 static ttfenc_entry *ttfenc_tab;
 static short *glyph_index;
-
+static int ncmaptabs;
+static cmap_entry *cmap_tab;
+static name_record *names_tab;
+static int names_record_num;
+static char *names_storage;
+static int names_storage_size;
 static tabdir_entry *dir_tab;
 static char *glyph_names_buf;
 static int chkcount;
-
 static FILE *ttf_file;
+static char id_str[] = "index";
+
 integer ttf_length;
 
 #define INFILE ttf_file
 #define OUTFILE pdffile
 
 #define TTF_OPEN()      truetypebopenin(ttf_file)
-#define TTF_CLOSE()     xfclose(ttf_file)
+#define TTF_CLOSE()     xfclose(ttf_file, filename)
 #define TTF_GETCHAR()   xgetc(ttf_file)
 #define TTF_EOF()       feof(ttf_file)
 
-char *mac_glyph_names[MAC_GLYPH_MAX] = {
-notdef, ".null", "CR", "space", "exclam", "quotedbl", "numbersign", "dollar",
-"percent", "ampersand", "quotesingle", "parenleft", "parenright", "asterisk",
-"plus", "comma", "hyphen", "period", "slash", "zero", "one", "two", "three",
-"four", "five", "six", "seven", "eight", "nine", "colon", "semicolon", "less",
-"equal", "greater", "question", "at", "A", "B", "C", "D", "E", "F", "G", "H",
-"I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W",
-"X", "Y", "Z", "bracketleft", "backslash", "bracketright", "asciicircum",
-"underscore", "grave", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k",
-"l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
-"braceleft", "bar", "braceright", "asciitilde", "Adieresis", "Aring",
-"Ccedilla", "Eacute", "Ntilde", "Odieresis", "Udieresis", "aacute", "agrave",
-"acircumflex", "adieresis", "atilde", "aring", "ccedilla", "eacute", "egrave",
-"ecircumflex", "edieresis", "iacute", "igrave", "icircumflex", "idieresis",
-"ntilde", "oacute", "ograve", "ocircumflex", "odieresis", "otilde", "uacute",
-"ugrave", "ucircumflex", "udieresis", "dagger", "degree", "cent", "sterling",
-"section", "bullet", "paragraph", "germandbls", "registered", "copyright",
-"trademark", "acute", "dieresis", "notequal", "AE", "Oslash", "infinity",
-"plusminus", "lessequal", "greaterequal", "yen", "mu", "partialdiff", "Sigma",
-"Pi", "pi", "integral", "ordfeminine", "ordmasculine", "Omega", "ae",
-"oslash", "questiondown", "exclamdown", "logicalnot", "radical", "florin",
-"approxequal", "Delta", "guillemotleft", "guillemotright", "ellipsis",
-"nbspace", "Agrave", "Atilde", "Otilde", "OE", "oe", "endash", "emdash",
-"quotedblleft", "quotedblright", "quoteleft", "quoteright", "divide",
-"lozenge", "ydieresis", "Ydieresis", "fraction", "currency", "guilsinglleft",
-"guilsinglright", "fi", "fl", "daggerdbl", "middot", "quotesinglbase",
-"quotedblbase", "perthousand", "Acircumflex", "Ecircumflex", "Aacute",
-"Edieresis", "Egrave", "Iacute", "Icircumflex", "Idieresis", "Igrave",
-"Oacute", "Ocircumflex", "applelogo", "Ograve", "Uacute", "Ucircumflex",
-"Ugrave", "dotlessi", "circumflex", "tilde", "overscore", "breve",
-"dotaccent", "ring", "cedilla", "hungarumlaut", "ogonek", "caron", "Lslash",
-"lslash", "Scaron", "scaron", "Zcaron", "zcaron", "brokenbar", "Eth", "eth",
-"Yacute", "yacute", "Thorn", "thorn", "minus", "multiply", "onesuperior",
-"twosuperior", "threesuperior", "onehalf", "onequarter", "threequarters",
-"franc", "Gbreve", "gbreve", "Idot", "Scedilla", "scedilla", "Cacute",
-"cacute", "Ccaron", "ccaron", "dmacron"
-};
+#include "macnames.c"
 
 static char *newtabnames[] = {
     "OS/2",
@@ -200,9 +186,9 @@ static char *newtabnames[] = {
 
 static unsigned char addchksm(unsigned char b)
 {
-    checksum += (b << (8*chkcount--));
-    if (chkcount < 0)
-        chkcount = 3;
+    checksum += (b << (8*(4 - ++chkcount)));
+    if (chkcount == 4)
+        chkcount = 0;
     return b;
 }
 
@@ -233,6 +219,13 @@ static long getnum(int s)
     return i;
 }
 
+static TTF_ULONG getchksm()
+{
+    if (chkcount != 0)
+        putnum(4 - chkcount, 0);
+    return checksum;
+}
+
 static long funit(long n)
 {
     if (n < 0)
@@ -290,15 +283,186 @@ static void copy_encoding()
     }
 }
 
-static void read_font()
+static void get_fontname()
 {
-    long i, j, k, l, n;
+    int i, l;
+    char *r, buf[1024], *p;
+    for (i = 0; i < names_record_num; i++)
+        if (names_tab[i].platform_id == 1 && 
+            names_tab[i].encoding_id == 0 &&
+            names_tab[i].name_id == 6) {
+            r = names_storage + names_tab[i].offset;
+            l = names_tab[i].length;
+            for (p = buf; l > 0; l--)
+                *p++ = *r++;
+            *p = 0;
+            font_keys[FONTNAME_CODE].value.s = xstrdup(buf);
+            font_keys[FONTNAME_CODE].valid = true;
+        }
+}
+
+static void read_name()
+{
+    int i;
+    tabdir_entry *tab = seek_tab("name", TTF_USHORT_SIZE);
+    char *p;
+    names_record_num = GET_USHORT();
+    names_tab = XTALLOC(names_record_num, name_record);
+    names_storage_size = tab->length - 
+         (3*TTF_USHORT_SIZE + names_record_num*6*TTF_USHORT_SIZE);
+    names_storage = XTALLOC(names_storage_size, char);
+    SKIP(TTF_USHORT_SIZE);
+    for (i = 0; i < names_record_num; i++) {
+        names_tab[i].platform_id = GET_USHORT();
+        names_tab[i].encoding_id = GET_USHORT();
+        names_tab[i].language_id = GET_USHORT();
+        names_tab[i].name_id = GET_USHORT();
+        names_tab[i].length = GET_USHORT();
+        names_tab[i].offset = GET_USHORT();
+    }
+    for (p = names_storage; p - names_storage < names_storage_size;)
+         *p++ = GET_CHAR();
+    get_fontname();
+}
+
+static void read_mapx()
+{
+    int i;
+    glyph_entry *glyph;
+    seek_tab("maxp", TTF_FIXED_SIZE);
+    glyph_tab = XTALLOC(1 + (glyphs_count = GET_USHORT()), glyph_entry);
+    for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count; glyph++) {
+        glyph->newindex = -1;
+        glyph->name_index = 0;
+        glyph->name = notdef;
+    }
+    glyph_index = XTALLOC(glyphs_count, short);
+    glyph_index[0] = 0; /* index of ".notdef" glyph */
+    glyph_index[1] = 1; /* index of ".null" glyph */
+    for (i = 0; i < MAX_KEY_CODE; i++)
+        if (i != FONTNAME_CODE)
+            font_keys[i].valid = true;
+}
+
+static void read_head()
+{
+    seek_tab("head", 2*TTF_FIXED_SIZE + 2*TTF_ULONG_SIZE + TTF_USHORT_SIZE);
+    upem = GET_USHORT();
+    SKIP(16);
+    font_keys[FONTBBOX1_CODE].value.i = funit(GET_FWORD());
+    font_keys[FONTBBOX2_CODE].value.i = funit(GET_FWORD());
+    font_keys[FONTBBOX3_CODE].value.i = funit(GET_FWORD());
+    font_keys[FONTBBOX4_CODE].value.i = funit(GET_FWORD());
+    SKIP(2*TTF_USHORT_SIZE + TTF_SHORT_SIZE);
+    loca_format = GET_SHORT();
+}
+
+static void read_hhea()
+{
+    seek_tab("hhea", TTF_FIXED_SIZE);
+    font_keys[ASCENT_CODE].value.i = funit(GET_FWORD());
+    font_keys[DESCENT_CODE].value.i = funit(GET_FWORD());
+    SKIP(TTF_FWORD_SIZE + TTF_UFWORD_SIZE + 3*TTF_FWORD_SIZE + 8*TTF_SHORT_SIZE);
+    nhmtxs = GET_USHORT();
+}
+
+static void read_pclt()
+{
+    if (name_lookup("PCLT", false) != 0) {
+        seek_tab("PCLT", TTF_FIXED_SIZE + TTF_ULONG_SIZE + TTF_USHORT_SIZE);
+        font_keys[XHEIGHT_CODE].value.i = funit(GET_USHORT());
+        SKIP(2*TTF_USHORT_SIZE);
+        font_keys[CAPHEIGHT_CODE].value.i = funit(GET_USHORT());
+    }
+    else {
+        font_keys[XHEIGHT_CODE].valid = false;
+        font_keys[CAPHEIGHT_CODE].valid = false;
+    }
+}
+
+static void read_hmtx()
+{
+    glyph_entry *glyph;
+    TTF_UFWORD last_advWidth;
+    seek_tab("hmtx", 0);
+    for (glyph = glyph_tab; glyph - glyph_tab < nhmtxs; glyph++) {
+        glyph->advWidth = GET_UFWORD();
+        glyph->lsb = GET_UFWORD();
+    }
+    if (nhmtxs < glyphs_count) {
+        last_advWidth = glyph[-1].advWidth;
+        for (;glyph - glyph_tab <  glyphs_count; glyph++) {
+            glyph->advWidth = last_advWidth;
+            glyph->lsb = GET_UFWORD();
+        }
+    }
+}
+
+static void read_post()
+{
+    int nnames;
+    long length;
     char *p;
     glyph_entry *glyph;
+    tabdir_entry *tab = seek_tab("post", 0);
+    post_format = GET_FIXED();
+    font_keys[ITALIC_ANGLE_CODE].value.i = GET_FIXED() >> 16;
+    SKIP(2*TTF_FWORD_SIZE + 5*TTF_ULONG_SIZE);
+    switch (post_format) {
+    case 0x10000:
+        for (glyph = glyph_tab; glyph - glyph_tab < NMACGLYPHS; glyph++) {
+            glyph->name = mac_glyph_names[glyph - glyph_tab];
+            glyph->name_index = glyph - glyph_tab;
+        }
+        break;
+    case 0x20000:
+        nnames = GET_USHORT(); /* some fonts have this value different from nglyphs */
+        for (glyph = glyph_tab; glyph - glyph_tab < nnames; glyph++)
+            glyph->name_index = GET_USHORT();
+        length = tab->length - (xftell(INFILE, filename) - tab->offset);
+        glyph_names_buf = XTALLOC(length, char);
+        for (p = glyph_names_buf; p - glyph_names_buf < length;) {
+            for (k = GET_BYTE(); k > 0; k--)
+                *p++ = GET_CHAR();
+            *p++ = 0;
+        }
+        for (glyph = glyph_tab; glyph - glyph_tab < nnames; glyph++) {
+            if (glyph->name_index < NMACGLYPHS)
+                glyph->name = mac_glyph_names[glyph->name_index];
+            else {
+                p = glyph_names_buf;
+                k = glyph->name_index - NMACGLYPHS;
+                for (; k > 0; k--)
+                    p = (char *)strchr(p, 0) + 1;
+                glyph->name = p;
+            }
+        }
+        break;
+    default:
+        FAIL("unsupported format (%8X) of `post' table"
+             AND (unsigned int) post_format);
+    }
+}
+
+static void read_loca()
+{
+    glyph_entry *glyph;
+    seek_tab("loca", 0);
+    if (loca_format != 0)
+        for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count + 1; glyph++)
+            glyph->offset = GET_ULONG();
+    else
+        for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count + 1; glyph++)
+            glyph->offset = GET_USHORT() << 1;
+}
+
+static void read_font()
+{
+    int i;
+    long version;
     tabdir_entry *tab;
-    TTF_UFWORD last_advWidth;
-    if ((i= GET_FIXED()) != 0x00010000)
-        FAIL("unsupport version %8lx; can handle only version 1.0" AND i);
+    if ((version= GET_FIXED()) != 0x00010000)
+        FAIL("unsupport version %8lx; can handle only version 1.0" AND version);
     dir_tab = XTALLOC(ntabs = GET_USHORT(), tabdir_entry);
     SKIP(3*TTF_USHORT_SIZE);
     for (tab = dir_tab; tab - dir_tab < ntabs; tab++) {
@@ -316,170 +480,153 @@ static void read_font()
         new_ntabs--;
     if (name_lookup("prep", false) == 0)
         new_ntabs--;
-    seek_tab("maxp", TTF_FIXED_SIZE);
-    glyph_tab = XTALLOC(1 + (glyphs_count = GET_USHORT()), glyph_entry);
-    for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count; glyph++)
-        glyph->newindex = -1;
-    glyph_index = XTALLOC(glyphs_count, short);
-    glyph_index[0] = 0; /* index of ".notdef" glyph */
-    glyph_index[1] = 1; /* index of ".null" glyph */
-    for (i = 0; i < MAX_KEY_CODE; i++)
-        font_keys[i].valid = true;
-    seek_tab("head", 2*TTF_FIXED_SIZE + 2*TTF_ULONG_SIZE + TTF_USHORT_SIZE);
-    upem = GET_USHORT();
-    SKIP(16);
-    font_keys[FONTBBOX1_CODE].value.i = funit(GET_FWORD());
-    font_keys[FONTBBOX2_CODE].value.i = funit(GET_FWORD());
-    font_keys[FONTBBOX3_CODE].value.i = funit(GET_FWORD());
-    font_keys[FONTBBOX4_CODE].value.i = funit(GET_FWORD());
-    SKIP(2*TTF_USHORT_SIZE + TTF_SHORT_SIZE);
-    loca_format = GET_SHORT();
-    seek_tab("hhea", TTF_FIXED_SIZE);
-    font_keys[ASCENT_CODE].value.i = funit(GET_FWORD());
-    font_keys[DESCENT_CODE].value.i = funit(GET_FWORD());
-    SKIP(TTF_FWORD_SIZE + TTF_UFWORD_SIZE + 3*TTF_FWORD_SIZE + 8*TTF_SHORT_SIZE);
-    nhmtxs = GET_USHORT();
-    if (name_lookup("PCLT", false) != 0) {
-        seek_tab("PCLT", TTF_FIXED_SIZE + TTF_ULONG_SIZE + TTF_USHORT_SIZE);
-        font_keys[XHEIGHT_CODE].value.i = funit(GET_USHORT());
-        SKIP(2*TTF_USHORT_SIZE);
-        font_keys[CAPHEIGHT_CODE].value.i = funit(GET_USHORT());
-    }
-    else {
-        font_keys[XHEIGHT_CODE].valid = false;
-        font_keys[CAPHEIGHT_CODE].valid = false;
-    }
-    seek_tab("hmtx", 0);
-    for (glyph = glyph_tab; glyph - glyph_tab < nhmtxs; glyph++) {
-        glyph->advWidth = GET_UFWORD();
-        glyph->lsb = GET_UFWORD();
-    }
-    if (nhmtxs < glyphs_count) {
-        last_advWidth = glyph[-1].advWidth;
-        for (;glyph - glyph_tab <  glyphs_count; glyph++) {
-            glyph->advWidth = last_advWidth;
-            glyph->lsb = GET_UFWORD();
-        }
-    }
-    tab = seek_tab("post", 0);
-    post_format = GET_FIXED();
-    font_keys[ITALIC_ANGLE_CODE].value.i = GET_FIXED() >> 16;
-    SKIP(2*TTF_FWORD_SIZE + 5*TTF_ULONG_SIZE);
-    switch (post_format) {
-    case 0x10000:
-        for (glyph = glyph_tab; glyph - glyph_tab < MAC_GLYPH_MAX; glyph++) {
-            glyph->name = mac_glyph_names[glyph - glyph_tab];
-            glyph->name_index = glyph - glyph_tab;
-        }
-        for (;glyph - glyph_tab < glyphs_count; glyph++) {
-            glyph->name = notdef;
-            glyph->name_index = 0;
-        }
-        break;
-    case 0x20000:
-        l = GET_USHORT(); /* some fonts have this value different from nglyphs */
-        for (glyph = glyph_tab; glyph - glyph_tab < l; glyph++)
-            glyph->name_index = GET_USHORT();
-        i = tab->length - (xftell(INFILE, filename) - tab->offset);
-        glyph_names_buf = XTALLOC(i, char);
-        for (p = glyph_names_buf; p - glyph_names_buf < i;) {
-            for (k = GET_BYTE(); k > 0; k--)
-                *p++ = GET_CHAR();
-            *p++ = 0;
-        }
-        for (glyph = glyph_tab; glyph - glyph_tab < l; glyph++) {
-            if (glyph->name_index < MAC_GLYPH_MAX)
-                glyph->name = mac_glyph_names[glyph->name_index];
-            else {
-                p = glyph_names_buf;
-                k = glyph->name_index - MAC_GLYPH_MAX;
-                for (; k > 0; k--)
-                    p = (char *)strchr(p, 0) + 1;
-                glyph->name = p;
-            }
-        }
-        break;
-    default:
-        FAIL("unsupported format (%8X) of `post' table" AND post_format);
-    }
-    seek_tab("loca", 0);
-    if (loca_format != 0)
-        for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count + 1; glyph++)
-            glyph->offset = GET_ULONG();
-    else
-        for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count + 1; glyph++)
-            glyph->offset = GET_USHORT() << 1;
-    seek_tab("name", TTF_USHORT_SIZE);
-    i = xftell(INFILE, filename);
-    n = GET_USHORT();
-    j = GET_USHORT() + i - TTF_USHORT_SIZE;
-    i += 2*TTF_USHORT_SIZE;
-    while (n-- > 0) {
-        seek_off("name", i + 3*TTF_USHORT_SIZE);
-        k = GET_USHORT();
-        l = GET_USHORT();
-        if (k == 6) {
-            seek_off("name", j + GET_USHORT());
-            CHECK_BUF(l, PRINTF_BUF_SIZE);
-            for (p = printf_buf; l > 0; l-- )
-                *p++ = GET_CHAR();
-            *p++ = 0;
-            font_keys[FONTNAME_CODE].value.s = xstrdup(printf_buf);
-            break;
-        }
-        i += 6*TTF_USHORT_SIZE;
-    }
+    read_mapx();
+    read_head();
+    read_hhea();
+    read_pclt();
+    read_hmtx();
+    read_post();
+    read_loca();
+    read_name();
 }
 
 #define RESET_CHKSM() do {                                 \
-    checksum = 0L;                                         \
-    chkcount = 3;                                          \
+    checksum = 0;                                          \
+    chkcount = 0;                                          \
     tab->offset = xftell(OUTFILE, filename) - ttf_offset;  \
 } while (0)
 
 #define SET_CHKSM() do {                                   \
-    tab->checksum = checksum;                               \
     tab->length = xftell(OUTFILE, filename) - tab->offset - ttf_offset; \
+    tab->checksum = getchksm();                            \
 } while (0)
 
 static void copytab(char *name)
 {
-    tabdir_entry *tab = seek_tab(name, 0);
     long i;
+    tabdir_entry *tab = seek_tab(name, 0);
     RESET_CHKSM();
     for (i = tab->length; i > 0; i--)
         COPY_CHAR();
     SET_CHKSM();
 }
 
-static void reenc_font()
+#define BYTE_ENCODING_LENGTH  (MAX_CHAR_NUM*TTF_BYTE_SIZE + 3*TTF_USHORT_SIZE)
+
+static void byte_encoding()
 {
     ttfenc_entry *e;
-    tabdir_entry *tab;
-    tab = name_lookup("cmap", true);
+    PUT_USHORT(0);  /* format number (0: byte encoding table) */
+    PUT_USHORT(BYTE_ENCODING_LENGTH); /* length of table */
+    PUT_USHORT(0);  /* version number */
+    for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++)
+        if (e->newindex < 256) {
+            PUT_BYTE(e->newindex);
+        }
+        else {
+            if (e->name != notdef)
+                WARN("glyph `%s' has been mapped to `%s' in `byte encoding' cmap table" AND
+                     e->name AND notdef);
+            PUT_BYTE(0); /* notdef */
+        }
+}
+
+#define SEG_MAP_DELTA_LENGTH ((16 + MAX_CHAR_NUM)*TTF_USHORT_SIZE)
+
+static void seg_map_delta()
+{
+    ttfenc_entry *e;
+    PUT_USHORT(4);  /* format number (4: segment mapping to delta values) */
+    PUT_USHORT(SEG_MAP_DELTA_LENGTH);
+    PUT_USHORT(0);  /* version number */
+    PUT_USHORT(4);  /* 2*segCount */
+    PUT_USHORT(4);  /* searchRange */
+    PUT_USHORT(1);  /* entrySelector */
+    PUT_USHORT(0);  /* rangeShift */
+    PUT_USHORT(255); /* endCount[0] */
+    PUT_USHORT(0xFFFF); /* endCount[1] */
+    PUT_USHORT(0); /* reversedPad */
+    PUT_USHORT(0); /* startCount[0] */
+    PUT_USHORT(0xFFFF); /* startCount[1] */
+    PUT_USHORT(0); /* idDelta[0] */
+    PUT_USHORT(1); /* idDelta[1] */
+    PUT_USHORT(2*TTF_USHORT_SIZE); /* idRangeOffset[0] */
+    PUT_USHORT(0); /* idRangeOffset[1] */
+    for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++)
+        PUT_USHORT(e->newindex);
+}
+
+#define CMAP_ENTRY_LENGTH (2*TTF_USHORT_SIZE + TTF_ULONG_SIZE)
+
+void static select_cmap()
+{
+    int i, n;
+    TTF_USHORT pid, eid, format;
+    TTF_ULONG offset, cur_offset, cmap_offset;
+    seek_tab("cmap", TTF_USHORT_SIZE); /* skip the table vesrion number (0) */
+    n = GET_USHORT();
+    cmap_tab = XTALLOC(n, cmap_entry);
+    cur_offset = xftell(INFILE, filename);
+    cmap_offset = cur_offset - 2*TTF_USHORT_SIZE;
+    ncmaptabs = 0;
+    for (i = 0; i < n; i++) {
+        xfseek(INFILE, cur_offset, SEEK_SET, filename);
+        pid = GET_USHORT();
+        eid = GET_USHORT();
+        offset = GET_ULONG();
+        cur_offset += CMAP_ENTRY_LENGTH;
+        xfseek(INFILE, cmap_offset + offset, SEEK_SET, filename);
+        format = GET_USHORT();
+        if (format == 0 || format == 4) {
+            cmap_tab[ncmaptabs].platform_id  = pid;
+            cmap_tab[ncmaptabs].encoding_id  = eid;
+            cmap_tab[ncmaptabs].offset       = offset;
+            cmap_tab[ncmaptabs].format       = format;
+            ncmaptabs++;
+        }
+        else
+            WARN("format %i of encoding subtable unsupported" AND (int)format);
+    }
+}
+
+static void write_cmap()
+{
+    cmap_entry *ce;
+    long offset;
+    tabdir_entry *tab = name_lookup("cmap", true);
+    select_cmap();
     RESET_CHKSM();
     PUT_USHORT(0);  /* table version number (0) */
-    PUT_USHORT(1);  /* number of encoding tables (1) */
-    PUT_USHORT(1);  /* platform ID */
-    PUT_USHORT(0);  /* encoding ID */
-    PUT_ULONG(4*TTF_USHORT_SIZE + TTF_ULONG_SIZE); /* byte offset from beginning of table */
-    if (max_glyph_index < 256) {
-        PUT_USHORT(0);  /* format number (0: byte encoding table) */
-        PUT_USHORT(256 + 3*TTF_USHORT_SIZE); /* length of table */
-        PUT_USHORT(0);  /* version number */
-        for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++)
-            PUT_BYTE(e->newindex);
+    PUT_USHORT(ncmaptabs);  /* number of encoding tables */
+    offset = 2*TTF_USHORT_SIZE + ncmaptabs*CMAP_ENTRY_LENGTH;
+    for (ce = cmap_tab; ce - cmap_tab < ncmaptabs; ce++) {
+        ce->offset = offset;
+        switch (ce->format) {
+        case 0: 
+            offset +=  BYTE_ENCODING_LENGTH;
+            break;
+        case 4: 
+            offset +=  SEG_MAP_DELTA_LENGTH;
+            break;
+        default:
+            FAIL("writting cmap: format %i unsupported" AND (int)ce->format);
+        }
+        PUT_USHORT(ce->platform_id);
+        PUT_USHORT(ce->encoding_id);
+        PUT_ULONG(ce->offset);
     }
-    else {
-        PUT_USHORT(6);  /* format number (6): trimmed table mapping */
-        PUT_USHORT(TTF_USHORT_SIZE*(5 + MAX_CHAR_NUM));
-        PUT_USHORT(0);  /* version number (0) */
-        PUT_USHORT(0);  /* first character code */
-        PUT_USHORT(MAX_CHAR_NUM);  /* number of character code in table */
-        for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++)
-            PUT_USHORT(e->newindex);
+    for (ce = cmap_tab; ce - cmap_tab < ncmaptabs; ce++) {
+        switch (ce->format) {
+        case 0: 
+            byte_encoding();
+            break;
+        case 4: 
+            seg_map_delta();
+            break;
+        }
     }
     SET_CHKSM();
+    XFREE(cmap_tab);
 }
 
 static void copy_dirtab()
@@ -510,68 +657,19 @@ static void copy_dirtab()
         }
 }
 
-
-static void subset_font()
+static void write_glyf()
 {
-    long i, k;
-    char *p;
-    TTF_USHORT flags;
+    short *id, k;
     TTF_USHORT idx;
-    tabdir_entry *tab;
-    short *index;
-    ttfenc_entry *e;
-    glyph_entry *glyph;
-    ttf_offset = xftell(OUTFILE, filename);
-    /* 
-     * reindexing glyphs: we append index of used glyphs to `glyph_index'
-     * while going through `ttfenc_tab'. After appending a new entry to
-     * `glyph_index' we set field `newindex' of corresponding entries in both
-     * `glyph_tab' and `ttfenc_tab' to the newly created index
-     * 
-     */
-    for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++) {
-        e->newindex = 0; /* index of ".notdef" glyph */
-        if (e->name != notdef) {
-            for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count; glyph++)
-                if (!strcmp(e->name, glyph->name))
-                    break;
-            if (glyph - glyph_tab < glyphs_count) {
-                glyph_index[new_glyphs_count] = glyph - glyph_tab;
-                glyph->newindex = e->newindex = new_glyphs_count;
-                new_glyphs_count++;
-            }
-            else
-                WARN("glyph `%s' undefined" AND e->name);
-        }
-    }
-    max_glyph_index = new_glyphs_count - 1;
-    for (i = 1, k = 0; i <= new_ntabs; i <<= 1, k++);
-    PUT_FIXED(0x00010000); /* font version */
-    PUT_USHORT(new_ntabs); /* number of tables */
-    PUT_USHORT(i << 3); /* search range */
-    PUT_USHORT(k - 1); /* entry selector */
-    PUT_USHORT((new_ntabs<<4) - (i<<3)); /* range shift */
-    xfseek(OUTFILE, ttf_offset + TABDIR_OFF + new_ntabs*4*TTF_ULONG_SIZE,
-        SEEK_SET, filename);
-    copytab("OS/2");
-    if (name_lookup("PCLT", false) != 0)
-        copytab("PCLT");
-    if (name_lookup("fpgm", false) != 0)
-        copytab("fpgm");
-    if (name_lookup("cvt ", false) != 0)
-        copytab("cvt ");
-    if (name_lookup("prep", false) != 0)
-        copytab("prep");
-    reenc_font();
-    i = (tab = name_lookup("glyf", true))->offset;
+    TTF_USHORT flags;
+    tabdir_entry *tab = name_lookup("glyf", true);
+    long glyf_offset = tab->offset;
+    long new_glyf_offset = xftell(OUTFILE, filename);
     RESET_CHKSM();
-    loca_format = 0;
-    for (index = glyph_index; index - glyph_index < new_glyphs_count; index++) {
-        if ((glyph_tab[*index].newoffset = xftell(OUTFILE, filename) - tab->offset - ttf_offset) > 0x00020000 ||
-            (glyph_tab[*index].newoffset & 1))
-            loca_format = 1;
-        if (glyph_tab[*index].offset != glyph_tab[*index + 1].offset) {
-            seek_off("glyph", i + glyph_tab[*index].offset);
+    for (id = glyph_index; id - glyph_index < new_glyphs_count; id++) {
+        glyph_tab[*id].newoffset = xftell(OUTFILE, filename) - new_glyf_offset;
+        if (glyph_tab[*id].offset != glyph_tab[*id + 1].offset) {
+            seek_off("glyph", glyf_offset + glyph_tab[*id].offset);
             k = COPY_SHORT();
             ncopy(4*TTF_FWORD_SIZE);
             if (k < 0) {
@@ -602,113 +700,204 @@ static void subset_font()
                     ncopy(COPY_USHORT());
             }
             else
-                ncopy(glyph_tab[*index + 1].offset - glyph_tab[*index].offset - 
+                ncopy(glyph_tab[*id + 1].offset - glyph_tab[*id].offset - 
                     TTF_USHORT_SIZE - 4*TTF_FWORD_SIZE);
         }
     }
+    last_glyf_offset = xftell(OUTFILE, filename) - new_glyf_offset;
     SET_CHKSM();
-    if ((i = tab->length) > 0x00020000 || i & 1)
-        loca_format = 1;
+}
+
+static void reindex_glyphs(boolean subset)
+{
+    ttfenc_entry *e;
+    glyph_entry *glyph;
+    int i, l = strlen(id_str);
+    /* 
+     * reindexing glyphs: we append index of used glyphs to `glyph_index'
+     * while going through `ttfenc_tab'. After appending a new entry to
+     * `glyph_index' we set field `newindex' of corresponding entries in both
+     * `glyph_tab' and `ttfenc_tab' to the newly created index
+     * 
+     */
+    for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++) {
+        e->newindex = 0; /* index of ".notdef" glyph */
+        if (e->name != notdef) {
+            if (strncmp(e->name, id_str, l) == 0) {
+                i = atoi(e->name + l);
+                if (i < glyphs_count)
+                    glyph = glyph_tab + i;
+                else
+                    glyph = glyph_tab + glyphs_count;
+            }
+            else {
+                for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count; glyph++)
+                    if (glyph->name != notdef && strcmp(glyph->name, e->name) == 0)
+                            break;
+            }
+            if (glyph - glyph_tab < glyphs_count) {
+                if (subset) {
+                    glyph_index[new_glyphs_count] = glyph - glyph_tab;
+                    glyph->newindex = e->newindex = new_glyphs_count;
+                    new_glyphs_count++;
+                }
+                else
+                    e->newindex = glyph - glyph_tab;
+            }
+            else
+                WARN("glyph `%s' not found" AND e->name);
+        }
+    }
+}
+ 
+static void write_head()
+{
+    tabdir_entry *tab;
     tab = seek_tab("head", 0);
     RESET_CHKSM();
     ncopy(2*TTF_FIXED_SIZE + 2*TTF_ULONG_SIZE + 2*TTF_USHORT_SIZE + 16 + 
         4*TTF_FWORD_SIZE + 2*TTF_USHORT_SIZE + TTF_SHORT_SIZE);
-    if (loca_format)
-        PUT_SHORT(1);
-    else
-        PUT_SHORT(0);
+    PUT_SHORT(loca_format);
     PUT_SHORT(0);
     SET_CHKSM();
+}
+ 
+static void write_hhea()
+{
+    tabdir_entry *tab;
     tab = seek_tab("hhea", 0);
     RESET_CHKSM();
     ncopy(TTF_FIXED_SIZE + 3*TTF_FWORD_SIZE + TTF_UFWORD_SIZE + 3*TTF_FWORD_SIZE + 8*TTF_SHORT_SIZE);
     PUT_USHORT(new_glyphs_count);
     SET_CHKSM();
-    tab = seek_tab("hmtx", 0);
+}
+
+static void write_htmx()
+{
+    short *id;
+    tabdir_entry *tab = seek_tab("hmtx", 0);
     RESET_CHKSM();
-    for (index = glyph_index; index - glyph_index < new_glyphs_count; index++) {
-        PUT_UFWORD(glyph_tab[*index].advWidth);
-        PUT_UFWORD(glyph_tab[*index].lsb);
+    for (id = glyph_index; id - glyph_index < new_glyphs_count; id++) {
+        PUT_UFWORD(glyph_tab[*id].advWidth);
+        PUT_UFWORD(glyph_tab[*id].lsb);
     }
     SET_CHKSM();
-    tab = seek_tab("loca", 0);
+}
+
+static void write_loca()
+{
+    short *id;
+    tabdir_entry *tab = seek_tab("loca", 0);
     RESET_CHKSM();
+    loca_format = 0;
+    if (last_glyf_offset >= 0x00020000 || (last_glyf_offset & 1))
+        loca_format = 1;
+    else
+        for (id = glyph_index; id - glyph_index < new_glyphs_count; id++)
+            if (glyph_tab[*id].newoffset & 1) {
+                loca_format = 1;
+                break;
+            }
     if (loca_format != 0) {
-        for (index = glyph_index; index - glyph_index < new_glyphs_count; index++)
-            PUT_ULONG(glyph_tab[*index].newoffset);
-        PUT_ULONG(i);
+        for (id = glyph_index; id - glyph_index < new_glyphs_count; id++)
+            PUT_ULONG(glyph_tab[*id].newoffset);
+        PUT_ULONG(last_glyf_offset);
     }
     else {
-        for (index = glyph_index; index - glyph_index < new_glyphs_count; index++)
-            PUT_USHORT(glyph_tab[*index].newoffset/2);
-        PUT_USHORT(i/2);
+        for (id = glyph_index; id - glyph_index < new_glyphs_count; id++)
+            PUT_USHORT(glyph_tab[*id].newoffset/2);
+        PUT_USHORT(last_glyf_offset/2);
     }
     SET_CHKSM();
-    tab = seek_tab("maxp", TTF_FIXED_SIZE + TTF_USHORT_SIZE);
+}
+
+static void write_mapx()
+{
+    tabdir_entry *tab = seek_tab("maxp", TTF_FIXED_SIZE + TTF_USHORT_SIZE);
     RESET_CHKSM();
     PUT_FIXED(0x00010000);
     PUT_USHORT(new_glyphs_count);
     ncopy(13*TTF_USHORT_SIZE);
     SET_CHKSM();
-    copytab("name");
-    tab = seek_tab("post", TTF_FIXED_SIZE);
+}
+
+static void write_post()
+{
+    int i;
+    char *p;
+    short *id;
+    tabdir_entry *tab = seek_tab("post", TTF_FIXED_SIZE);
     RESET_CHKSM();
     PUT_FIXED(0x00020000);
     ncopy(TTF_FIXED_SIZE + 2*TTF_FWORD_SIZE + 5*TTF_ULONG_SIZE);
     PUT_USHORT(new_glyphs_count);
-    for (k = 0, index = glyph_index; index - glyph_index < new_glyphs_count; index++)
-        if (glyph_tab[*index].name_index < MAC_GLYPH_MAX)
-            PUT_USHORT(glyph_tab[*index].name_index);
-        else
-            PUT_USHORT(MAC_GLYPH_MAX + k++);
-    for (index = glyph_index; index - glyph_index < new_glyphs_count; index++)
-        if (glyph_tab[*index].name_index >= MAC_GLYPH_MAX) {
-            PUT_BYTE(k = strlen(p = glyph_tab[*index].name));
-            while (k-- > 0)
+    for (k = 0, id = glyph_index; id - glyph_index < new_glyphs_count; id++) {
+        i = glyph_tab[*id].name_index;
+        if (i >= NMACGLYPHS)
+            i = NMACGLYPHS + k++;
+        PUT_USHORT(i);
+    }
+    for (id = glyph_index; id - glyph_index < new_glyphs_count; id++)
+        if (glyph_tab[*id].name_index >= NMACGLYPHS) {
+            p = glyph_tab[*id].name;
+            i = strlen(p);
+            PUT_BYTE(i);
+            while (i-- > 0)
                 PUT_CHAR(*p++);
         }
     SET_CHKSM();
+}
+
+static void init_font(int n)
+{
+    int i, k;
+    ttf_offset = xftell(OUTFILE, filename);
+    for (i = 1, k = 0; i <= new_ntabs; i <<= 1, k++);
+    PUT_FIXED(0x00010000); /* font version */
+    PUT_USHORT(n); /* number of tables */
+    PUT_USHORT(i << 3); /* search range */
+    PUT_USHORT(k - 1); /* entry selector */
+    PUT_USHORT((n<<4) - (i<<3)); /* range shift */
+    xfseek(OUTFILE, ttf_offset + TABDIR_OFF + n*4*TTF_ULONG_SIZE,
+        SEEK_SET, filename);
+}
+
+static void subset_font()
+{
+    init_font(new_ntabs);
+    copytab("OS/2");
+    if (name_lookup("PCLT", false) != 0)
+        copytab("PCLT");
+    if (name_lookup("fpgm", false) != 0)
+        copytab("fpgm");
+    if (name_lookup("cvt ", false) != 0)
+        copytab("cvt ");
+    if (name_lookup("prep", false) != 0)
+        copytab("prep");
+    reindex_glyphs(true);
+    write_glyf();
+    write_loca();
+    write_head();
+    write_hhea();
+    write_htmx();
+    write_mapx();
+    copytab("name");
+    write_post();
+    write_cmap();
     copy_dirtab();
 }
 
 static void copy_font()
 {
-    ttfenc_entry *e;
     tabdir_entry *tab;
-    glyph_entry *glyph;
-    int i;
-    ttf_offset = xftell(OUTFILE, filename);
-    for (i = 1, k = 0; i <= ntabs; i <<= 1, k++);
-    PUT_FIXED(0x00010000); /* font version */
-    PUT_USHORT(ntabs); /* number of tables */
-    PUT_USHORT(i << 3); /* search range */
-    PUT_USHORT(k - 1); /* entry selector */
-    PUT_USHORT((ntabs<<4) - (i<<3)); /* range shift */
-    xfseek(OUTFILE, ttf_offset + TABDIR_OFF + ntabs*4*TTF_ULONG_SIZE,
-        SEEK_SET, filename);
-    for (i = 0; i < ntabs; i++) {
-        if (!is_reencoded() || (strncmp(dir_tab[i].tag, "cmap", 4) != 0)) {
-            copytab(dir_tab[i].tag);
-            continue;
+    init_font(ntabs);
+    for (tab = dir_tab; tab - dir_tab < ntabs; tab++)
+        if (is_reencoded() && (strncmp(tab->tag, "cmap", 4) == 0)) {
+            reindex_glyphs(false);
+            write_cmap();
         }
-        tab = name_lookup("cmap", true);
-        max_glyph_index = 0;
-        for (e = ttfenc_tab; e - ttfenc_tab < MAX_CHAR_NUM; e++) {
-            e->newindex = 0; /* index of ".notdef" glyph */
-            if (e->name != notdef) {
-                for (glyph = glyph_tab; glyph - glyph_tab < glyphs_count; glyph++)
-                    if (strcmp(e->name, glyph->name) == 0)
-                        break;
-                if (glyph - glyph_tab < glyphs_count)
-                    e->newindex = glyph - glyph_tab;
-                else
-                    WARN("glyph `%s' undefined" AND e->name);
-                if (e->newindex > max_glyph_index)
-                    max_glyph_index = e->newindex;
-            }
-        }
-        reenc_font();
-    }
+        else
+            copytab(tab->tag);
     copy_dirtab();
 }
 
@@ -729,23 +918,35 @@ void writettf()
     glyph_names_buf = 0;
     new_glyphs_count = 2;
     new_ntabs = DEFAULT_NTABS;
+    dir_tab = 0;
+    glyph_tab = 0;
+    glyph_index = 0;
+    glyph_names_buf = 0;
+    names_tab = 0;
+    names_storage = 0;
+    ttfenc_tab = 0;
     read_font();
     if (is_included()) {
-        pdfsaveoffset = pdfoffset;
+        pdfsaveoffset = pdfoffset();
         pdfflush();
-        copy_encoding();
-        if (!is_subsetted())
-            copy_font();
-        else
+        if (is_reencoded())
+            copy_encoding();
+        if (is_subsetted())
             subset_font();
+        else
+            copy_font();
         xfflush(OUTFILE);
         xfseek(OUTFILE, pdfgone, SEEK_SET, filename);
-        ttf_length = pdfoffset - pdfsaveoffset;
-        XFREE(ttfenc_tab);
+        ttf_length = pdfoffset() - pdfsaveoffset;
+        if (is_reencoded())
+            XFREE(ttfenc_tab);
     }
     XFREE(dir_tab);
     XFREE(glyph_tab);
     XFREE(glyph_index);
     XFREE(glyph_names_buf);
+    XFREE(names_tab);
+    XFREE(names_storage);
+    TTF_CLOSE();
     TEX_PRINTF(">");
 }
